/*
 *	Arpscan
 *
 *	Copyright (C) 2002 Jan Bobrowski <jb@wizard.ae.krakow.pl>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	version 2 as published by the Free Software Foundation.
 */

#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <poll.h>
#include <errno.h>
#include <err.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <arpa/inet.h>

#define _STR(S) #S
#define STR(S) _STR(S)

#define ARP htons(ETHERTYPE_ARP)
#define IP htons(ETHERTYPE_IP)

typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;

#include "list.h"
#define elemof(T) (sizeof T/sizeof*T)
#define endof(T) (T+elemof(T))

#define HWMAX 8

int sock; /* packet socket */
struct sockaddr_ll bcast;
int nosort = 0, noown = 0, noethn = 0, proui = 0;

void print_oui(u8 a[6]);

struct he;
void print_he(struct he *he);

struct ifinfo {
	int index;
	char *name;
	u32 ip, net, mask, bcast;
	u16 hw_type;
	struct hwaddr {
		u8 len, addr[HWMAX];
	} hw;
} ifinfo;

static inline innet(u32 ip)
{
	return (ip&ifinfo.mask) == ifinfo.net;
}

static inline u32 ip_from_sa(struct sockaddr *sa)
{
	return ntohl(((struct sockaddr_in*)sa)->sin_addr.s_addr);
}

/* TABLE */

struct list hashtbl[64];
struct he {
	struct list hash;
	u32 ip;
	struct hwaddr hw;
};

static void init_hash() __attribute__((constructor));
static void init_hash()
{
	int i;
	for(i=0;i<elemof(hashtbl);i++) list_init(&hashtbl[i]);
}

struct list *list_for(u32 ip)
{
	return &hashtbl[ip&elemof(hashtbl)-1];
}

struct he *add_he(u32 ip, u8 *ha, u8 hl)
{
	struct he *he = (struct he*)malloc(sizeof *he);
	struct list *h = list_for(ip), *l;
	he->ip = ip;
	he->hw.len = hl;
	memcpy(he->hw.addr, ha, hl);
	for(l=h->next; l!=h; l=l->next) {
		struct he *he2 = list_entry(l, struct he, hash);
		if(he->ip < he2->ip) break;
	}
	list_add(l->prev, &he->hash);
	if(nosort) print_he(he);
	return he;
}

struct he *find_he(u32 ip)
{
	struct list *l, *h = list_for(ip);
	for(l=h->next; l!=h; l=l->next) {
		struct he *he = list_entry(l, struct he, hash);
		if(he->ip == ip) return he;
	}
	return 0;
}

/* INTERFACE */

static int net;

static void my__ioctl(int i, struct ifreq *r, char *t)
{
	if(ioctl(net, i, r) < 0)
		err(1, "ioctl(%s,%s)", t, r->ifr_name);
}
#define my_ioctl(I,R) my__ioctl(I,R,#I)

void fill_ifinfo(char *name)
{
	struct ifreq ir;
	int flags;

	ifinfo.index = if_nametoindex(name);
	if(!ifinfo.index) errx(1, "No such interface: %s", name);
	ifinfo.name = name;

	net = socket(PF_INET, SOCK_DGRAM, 0);
	if(net<0) err(1, "socket(PF_INET)");
	strcpy(ir.ifr_name, ifinfo.name);
	my_ioctl(SIOCGIFFLAGS, &ir);
	flags = ir.ifr_flags;
	if(flags & IFF_NOARP) errx(1, "%s: ARP not supported.", name);
	my_ioctl(SIOCGIFADDR, &ir);
	ifinfo.ip = ip_from_sa(&ir.ifr_addr);
	if(flags & IFF_POINTOPOINT) {
		my_ioctl(SIOCGIFDSTADDR, &ir);
		ifinfo.net = ip_from_sa(&ir.ifr_dstaddr);
		ifinfo.mask = ~(u32)0;
		ifinfo.bcast = 0; /* none */
	} else {
		my_ioctl(SIOCGIFNETMASK, &ir);
		ifinfo.mask = ip_from_sa(&ir.ifr_netmask);
		my_ioctl(SIOCGIFBRDADDR, &ir);
		ifinfo.bcast = ip_from_sa(&ir.ifr_broadaddr);
		ifinfo.net = ifinfo.ip & ifinfo.mask;
	}
	close(net);
}

static inline char *str_ip(u32 ip)
{
	struct in_addr n;
	n.s_addr = htonl(ip);
	return inet_ntoa(n);
}

char *str_hw(u8 *a, int l)
{
	static char buf[3*HWMAX];
	char *d = buf;
	if(l) for(;;) {
		d += sprintf(d, "%02X", *a++);
		if(--l <= 0) break;
		*d++ = ':';
	}
	*d = 0;
	return buf;
}

/* SCAN */

struct arppkt {
	u16 hrd, pro;
	u8 hln, pln;
	u16 op;
	u8 a[2*HWMAX+2*4];
/*	u8 sha[6];
	u8 sip[4];
	u8 tha[6];
	u8 tip[4];*/
};

#if 0
void print_arp(struct arppkt *arp)
{
	u8 *p = arp->a;
	printf("hrd:%04X pro:%04X ", ntohs(arp->hrd), ntohs(arp->pro));
	printf("hln:%d pln:%d op:%d ", arp->hln, arp->pln, ntohs(arp->op));
	printf("sha:%s ", str_hw(p, arp->hln)); p+=arp->hln;
	printf("sip:%s ", str_ip(ntohl(*(u32*)p))); p+=arp->pln;
	printf("tha:%s ", str_hw(p, arp->hln)); p+=arp->hln;
	printf("tip:%s\n", str_ip(ntohl(*(u32*)p)));
}
#endif

u32 scanip;
int scann = 6;

sendscan()
{
	struct arppkt arp;
	int ns;
	u8 *p;

	arp.hrd = htons(ifinfo.hw_type);
	arp.pro = IP;
	arp.hln = ifinfo.hw.len;
	arp.pln = 4;
	arp.op = htons(1);
	p = arp.a;
	memcpy(p, ifinfo.hw.addr, ifinfo.hw.len); p += ifinfo.hw.len;
	*(u32*)p = htonl(ifinfo.ip); p += 4;
	memset(p, 0, ifinfo.hw.len); p += ifinfo.hw.len;

	ns = 0;
	while(innet(scanip)) {
		int v;
		if(scanip == ifinfo.bcast || find_he(scanip)) {
			scanip++;
			continue;
		}
		*(u32*)p = htonl(scanip);
		v = sendto(sock, &arp, p+4-(u8*)&arp, 0, (struct sockaddr*)&bcast, sizeof bcast);
		if(v<0) {
			if(errno != ENOBUFS || scann <= 1)
				err(1, "send(%s)", str_hw(bcast.sll_addr, bcast.sll_halen));
			scann--;
			return -1;
		}
		scanip++;
		if(++ns >= scann) break;
	}
	return ns;
}

void receive()
{
	struct sockaddr_ll addr;
	struct arppkt arp;
	struct he *he;
	u32 ip;
	int v;

	{
		socklen_t l = sizeof addr;
		v = recvfrom(sock, &arp, sizeof arp, 0, (struct sockaddr*)&addr, &l);
		if(v < 0) err(1, "recvfrom");
	}
	if(v < arp.a-(u8*)&arp)
		return;
	if(arp.op != htons(2)) /* only responses */
		return;
	if(arp.pro != IP)
		return;
	if(arp.hrd != htons(ifinfo.hw_type) || arp.hln != ifinfo.hw.len)
		return;
	if(v < arp.a-(u8*)&arp + 2*arp.hln + 2*4)
		return;

	ip = ntohl(*(u32*)(arp.a+arp.hln));
	if(addr.sll_halen != arp.hln || memcmp(addr.sll_addr, arp.a, arp.hln)) {
		fprintf(stderr, "%s", str_hw(addr.sll_addr, addr.sll_halen));
		fprintf(stderr, " reported %s %s\n", str_hw(arp.a, arp.hln), str_ip(ip));
	}

	he = find_he(ip);
	if(!he)
		add_he(ip, arp.a, arp.hln);
	else {
		if(he->hw.len != arp.hln || memcmp(he->hw.addr, arp.a, arp.hln)) {
			fprintf(stderr, "%s", str_hw(he->hw.addr, he->hw.len));
			fprintf(stderr, " != %s\n", str_hw(arp.a, arp.hln));
		}
	}
}

/**/

int waitsock(int n)
{
	int v;
	struct pollfd pollfd;
	pollfd.fd = sock;
	pollfd.events = POLLIN;
	v = poll(&pollfd, 1, n);
	if(v < 0) {
		if(errno != EINTR)
			err(1, "poll");
		v = 0;
	}
	return v;
}

void print_he(struct he *he)
{
	if(noown && he->ip == ifinfo.ip) return;
	printf("%s %-15s", str_hw(he->hw.addr, he->hw.len), str_ip(he->ip));
	if(ifinfo.hw_type == ARPHRD_ETHER) {
		if(proui)
			print_oui(((struct ether_addr*)he->hw.addr)->ether_addr_octet);
		else if(!noethn) {
			char nm[256];
			if(!ether_ntohost(nm, (struct ether_addr*)he->hw.addr))
				printf(" %s", nm);
		}
	}
	putchar('\n');
}

int main(int argc, char **argv)
{
	{
		int o;
		while((o=getopt(argc, argv, "faepV?"))!=EOF) switch(o) {
			case 'f': nosort=1; break;
			case 'a': noown=1; break;
			case 'e': noethn=1; break;
			case 'p': proui=1; break;
			case '?': printf(
				"arpscan [-faepV] [interface]\n"
				"\t-f  do not sort responses\n"
				"\t-a  do not list interface's own address\n"
				"\t-e  do not include info from /etc/ethers\n"
				"\t-p  print vendor names\n"
				"\t-V  version info\n"
				);
				return 0;
			case 'V':
				printf("Arpscan by Jan Bobrowski 2002\n"
					"Version: " STR(VER) " (compiled: " __DATE__")\n"
					"License: GPL\n");
				return 0;
		}
	}

	fill_ifinfo(optind<argc ? argv[optind] : "eth0");

	{
		struct sockaddr_ll addr = {0};
		int l;

		sock = socket(PF_PACKET, SOCK_DGRAM, 0);
		if(sock < 0) err(1, "socket(AF_PACKET)");

		addr.sll_family = AF_PACKET;
		addr.sll_protocol = ARP;
		addr.sll_ifindex = ifinfo.index;
	
		if(bind(sock, (struct sockaddr*)&addr, sizeof addr)<0)
			err(1, "bind");
		l = sizeof addr;
		if(getsockname(sock, (struct sockaddr*)&addr, &l)<0)
			err(1, "getsockname");

		if(addr.sll_halen > HWMAX)
			errx(1, "hardware address too long (%d)", addr.sll_halen);
		ifinfo.hw.len = addr.sll_halen;
		memcpy(ifinfo.hw.addr, addr.sll_addr, sizeof ifinfo.hw.addr);
		ifinfo.hw_type = addr.sll_hatype;
	}

	/* hw broadcast address is Linux's secret, this works with Ethernet */
	bcast.sll_family = AF_PACKET;
	bcast.sll_protocol = ARP;
	bcast.sll_ifindex = ifinfo.index;
	bcast.sll_hatype = ifinfo.hw_type;
	bcast.sll_pkttype = PACKET_BROADCAST; /* unused :-( */
	bcast.sll_halen = ifinfo.hw.len;
	memset(bcast.sll_addr, 0xFF, ifinfo.hw.len);

	add_he(ifinfo.ip, ifinfo.hw.addr, ifinfo.hw.len);

	scanip = ifinfo.net;
	while(sendscan()) {
		while(waitsock(10))
			receive();
	}
	scanip = ifinfo.net;
	while(sendscan()) {
		while(waitsock(10))
			receive();
	}
	while(waitsock(100))
		receive();

	if(!nosort) for(scanip = ifinfo.net; innet(scanip); scanip++) {
		struct he *he = find_he(scanip);
		if(he)
			print_he(he);
	}
}
